#include <asm/io.h>
#include <stdio.h>

char porto;

#define pausa1 20 // pause between one bit and the next bit
#define pausa2 10000 // pause for reset
#define pausa3 10000 // pause after an erase cycle
#define pausa4 100 // pause after a comand or data



/* The file PICPORT.C contains all the hardware-dependant functions. If you
 * want to add support for other PIC programmer, just modify that file, and
 * leave this untouched. */

#include "picport.c"


/* This function prints an error given in MESSAGE, closes all the ports opened
 * and exits */

void error(char *message) {
  printf("%s\n",message);
  ending();
  exit(1);
}

/* This function sends a 6-bit command to the PIC, in the right 6-bit format
 * described in the Microchip's datasheets. */

void envia_comando(char comando) {
  char bit;
  int contador;

  for(contador=0;contador<6;contador++) {
    bit=comando&1; // bit inferior
    envia_bit(bit);
    comando/=2;
  }
  publicidad(pausa4);
}

/* This function sends a 14-bit data to the PIC, in the right 16-bit length
 * format described in the Microchip's datasheets. */

void envia_dato(int dato) {
  char bit;
  int contador;

  envia_bit(0);
  for(contador=0;contador<14;contador++) {
    bit=(char) (dato&1); // bit inferior
    envia_bit(bit);
    dato/=2;
  }
  envia_bit(0);
  publicidad(pausa4);
}

/* This function erases the program and data memory of the PIC. If it receives
 * a 0, it performs an erase cycle for PIC16C84. If not, if performs an erase
 * cycle for PIC16F84. The main program will use both cycles to ensure that all
 * PICs are conveniently erased */

void borrar(char tipo) {

  if(tipo==0) {
    envia_comando(2);
    envia_dato(0x3FFF);
    envia_comando(9);
    envia_comando(8);
    usleep(pausa3);
  } else {
    envia_comando(2);
    envia_dato(0x3FFF);
    envia_comando(1);
    envia_comando(7);
    envia_comando(8);
    usleep(pausa3);
    envia_comando(1);
    envia_comando(7);
  }
}

/* This function allows to set the CONFIGURATION register of the PIC. If the
 * protection bits are unset, it can change all the word. If not, only the
 * four less significant bits can be changed. To unprotect a protected PIC,
 * you must use PROTEGE instead of PROTEGE2. But warning, because PROTEGE
 * will erase the whole memory of the PIC */

void protege2 (int codigo) {
  int contador;

  printf("Protect with %d ",codigo);
  envia_comando(0); // Load Configuration
  envia_dato(codigo);
  for(contador=0;contador<7;contador++)
    envia_comando(6); // Increment Address (paso a configuration word)
  envia_comando(8);
  usleep(pausa3);
  envia_comando(4);
  contador=leer_dato();
  printf("Configuration readed: %d\n",contador);
  usleep(pausa4);
}

/* This function unprotects the PIC. Is like PROTEGE2, but can set and
 * unset the whole bits of the CONFIGURATION register even if the protection
 * is set (of course, in this case, the PIC memory is automatically erased) */

void protege (int codigo) {
  int contador;

  printf("Unprotecting the PIC.\n");
  envia_comando(0); // Load Configuration
  envia_dato(codigo);
  for(contador=0;contador<7;contador++)
    envia_comando(6); // Increment Address (paso a configuration word)
  envia_comando(1);
  envia_comando(7);
  envia_comando(8);
  usleep(pausa3);
  envia_comando(4);
  contador=leer_dato();
  printf("Configuration readed: %d\n",contador);
  if(contador!=codigo)
    error("Error when unprotecting the PIC.");
  usleep(pausa4);
}

/* This function reads a word from the PIC */

int leer_dato(void) {
  int dato,contador;
  char temporal;

  dato=0;
  lee_bit(); // First, we read the start bit
  for(contador=0;contador<14;contador++) {
    dato=dato/2; // Here, we rotate the data one possition to the right
    temporal=lee_bit(); // Here, we read the bit
    if(temporal!=0) // if the bit is different of one, we add 8192
      dato+=8192; // to the data (wich will be rotated in the division)
  }
  lee_bit(); // Here, we read the stop bit
  publicidad(pausa4); // And wait the pause between commands.
  return dato;
}

/* MAIN function */

int main(int argc,char *argv[]) {
  int basura,contador,proteccion,temporal,comando;
  char estado,memoria[10],nombre[256],lectura;
  FILE *fichero;


  porto=0;
  lectura=0;
  nombre[0]=0;
  for(contador=1;contador<argc;contador++) {
    if(strcmp(argv[contador],"-p")==0)
      if(contador==(argc-1)) {
	printf("You forgot to write the port number.\n");
	exit(1);
      } else {
	contador++;
	porto=argv[contador][0]-48;
      }
    else
      if(strcmp(argv[contador],"-r")==0)
	lectura=1;
      else
	strcpy(nombre,argv[contador]);
  }

  if(nombre[0]==0) {
    printf("You forgot to write the file name.\n");
    exit(1);
  }

  if(lectura==0)
    fichero=fopen(nombre,"r");
  else
    fichero=fopen(nombre,"w");
  if(fichero==NULL) {
    printf("I can't open the file %s.\n",nombre);
    exit(1);
  }

  basura=init();
  if(basura!=0) {
    fclose(fichero);
    if(basura==1)
      printf("The specified port is not supported.\n");
    if(basura==2)
      printf("IOPERM call failed. Be sure that you are executing\nthis program with ROOT priviledges.\n");
    if(basura==3)
      printf("OPEN call failed. Be sure that the /dev/ttySn device has\nthe right permissions.\n");
    exit(1);
  }

  reset(); // Setting up the PIC

  if(lectura==0) { // I want to write a new program in the PIC

    /* Here I unprotect the PIC, allowing the system to reprogram it */
    
    printf("\nUnprotecting the PIC\n");
    
    protege(0x3FFF);
    reset();
    
    /* Here, I erase the PIC memory */
    
    printf("Erasing the PIC\n");
    
    borrar(1);
    reset();
    borrar(0);
    reset();
    
    /* Here I read the protection word for the PIC, wich is in the beginning
     * of the .OBJ file */
    
    estado=0;
    basura=fscanf(fichero,"%x",&proteccion);
    printf("Configuration word: %d \n",proteccion);
    printf("Sending code\n");
    while(!feof(fichero)) {
      temporal=-1;
      memoria[0]=0;
      basura=fscanf(fichero,"%s",memoria);
      if(strcmp(memoria,"data:")==0) {
	estado=1;
	reset();
	printf("Sending EEPROM data\n");
      } else {
	basura=sscanf(memoria,"%X",&temporal);
	if(temporal>=0) {
	  printf("Data: %d ",temporal);
	  if(estado==0)
	    comando=2; // Load data for Program Memory
	  else
	    comando=3; // Load data for Data Memory
	  envia_comando(comando);
	  envia_dato(temporal);
	  envia_comando(8); // since the command 8 includes an erase cycle, we
	  publicidad(pausa3); // must wait 10 milliseconds at least.
	  if(estado==0)
	    comando=4; // Read data for Program Memory
	  else
	    comando=5; // Read data for Data Memory
	  envia_comando(comando);
	  basura=leer_dato();
	  if(estado==1)
	    basura=basura&0xFF;
	  printf("Readed: %d \n",basura);
	  if(basura!=temporal)
	    error("Error while verifying the data stored. PIC damaged? Programmer turned off?\n(Maybe a bug in my program...? glups)\n");
	  envia_comando(6);
	}
      }
    }
    
    protege2(proteccion);
    printf("\nProgramming cycle ended. Enjoy!!!!!\n\n");

  } else { // If not, I want to read the data in the PIC
    envia_comando(0);
    envia_dato(0); // filling data
    for(contador=0;contador<7;contador++)
      envia_comando(6);
    envia_comando(4);
    basura=leer_dato();
    fprintf(fichero,"%X\n",basura);
    printf("CONFIGURATION register: %X\n",basura);
    reset();
    printf("Code memory:\n");
    for(contador=0;contador<1024;contador++) {
      envia_comando(4);
      basura=leer_dato();
      envia_comando(6);
      fprintf(fichero,"%X\n",basura);
      printf("%X\t",basura);
      if(contador%10==9) {
	printf("\n");
	fflush(NULL);
      }
    }
    reset();
    printf("\nData memory:\n");
    fprintf(fichero,"data:\n");
    for(contador=0;contador<64;contador++) {
      envia_comando(5);
      basura=leer_dato();
      envia_comando(6);
      basura=basura&0xFF;
      fprintf(fichero,"%X\n",basura);
      printf("%X\t",basura);
      if(contador%10==9) {
	printf("\n");
	fflush(NULL);
      }
    }
    printf("\n");
  }
  ending();
  fclose(fichero);
}
